#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <assert.h>
#include <typeinfo>
#include <map>

using namespace std;
typedef int64_t sizeType;
typedef double scalar;
#define ASSERT(A) assert(A);
#define ASSERT_MSG(A,MSG) if(!(A))cout << MSG << endl;

//data structure
struct QuadraticCurve
{
  //_coef[0] = gravity
  //_coef[1] = velocity
  //_coef[2] = position
  scalar _coef[3][3];
};
struct QuadraticCurveSet
{
  vector<QuadraticCurve> _curves;
  sizeType _nrCandidate;  //number of particles leaving the source container at this frame
  scalar _time0; //current physical time
};
struct FluidRLData
{
  vector<QuadraticCurveSet> _curves;  //a set of quadratic curves drawn by particles leaving the source container
  vector<vector<scalar> > _features,_actions; //the liquid/rigid features of source container
};

//visualizer
class Endianness
{
public:
  static bool isLittleEndian() {
    union u {
      unsigned long l;
      unsigned char c[sizeof(unsigned long)];
    };
    u dummy;
    dummy.l = 1;
    return dummy.c[0] == 1;
  }
  static void swap2Bytes(unsigned char* &ptr) {
    unsigned char tmp;
    tmp = ptr[0];
    ptr[0] = ptr[1];
    ptr[1] = tmp;
  }
  static void swap4Bytes(unsigned char* &ptr) {
    unsigned char tmp;
    tmp = ptr[0];
    ptr[0] = ptr[3];
    ptr[3] = tmp;
    tmp = ptr[1];
    ptr[1] = ptr[2];
    ptr[2] = tmp;
  }
  static void swap8Bytes(unsigned char* &ptr) {
    unsigned char tmp;
    tmp = ptr[0];
    ptr[0] = ptr[7];
    ptr[7] = tmp;
    tmp = ptr[1];
    ptr[1] = ptr[6];
    ptr[6] = tmp;
    tmp = ptr[2];
    ptr[2] = ptr[5];
    ptr[5] = tmp;
    tmp = ptr[3];
    ptr[3] = ptr[4];
    ptr[4] = tmp;
  }
};
template<class T2>
void byteSwap(T2& val)
{
  sizeType n=sizeof(T2);
  unsigned char *p=(unsigned char*)&val;
  switch( n ) {
  case 1:
    return;
  case 2:
    Endianness::swap2Bytes(p);
    break;
  case 4:
    Endianness::swap4Bytes(p);
    break;
  case 8:
    Endianness::swap8Bytes(p);
    break;
  default:
    break;
  }
}
template <typename T2>
void vtkWriteTpl(ostream& oss,T2 val)
{
  if(typeid(val) == typeid(double)) {
    double valD=(double)(val);
    if(Endianness::isLittleEndian())
      byteSwap(valD);
    oss.write((const char*)&valD,sizeof(double));
  } else {
    if(Endianness::isLittleEndian())
      byteSwap(val);
    oss.write((const char*)&val,sizeof(T2));
  }
}
void vtkWrite(ostream& oss,float val)
{
  vtkWriteTpl(oss,val);
}
void vtkWrite(ostream& oss,double val)
{
  vtkWriteTpl(oss,val);
}
void vtkWrite(ostream& oss,int val)
{
  vtkWriteTpl(oss,val);
}
void vtkWrite(ostream& oss,unsigned char val)
{
  oss.write((const char*)&val,1);
}
template<typename T>
struct VTKWriter {
  enum VTK_DATA_TYPE {
    UNSTRUCTURED_GRID,
    STRUCTURED_POINTS,
  };
  enum VTK_CELL_TYPE {
    POINT=1,
    LINE=3,
    TRIANGLE=5,
    TETRA=10,
    VOXEL=11,
    HEX=12,
    POLYLINE=4,
    QUAD=9,
    QUADRATIC_LINE=21,
  };
public:
  struct Data {
    Data():_nr(0) {}
    string _str;
    sizeType _nr;
  };
  template <typename ITER>
  struct ValueTraits {
    typedef typename ITER::value_type value_type;
  };
  template <typename POINTED_TYPE>
  struct ValueTraits<POINTED_TYPE*> {
    typedef POINTED_TYPE value_type;
  };
  template <typename ID>
  struct IteratorIndex {
    typedef ID value_type;
    IteratorIndex(const sizeType& id,const sizeType stride,const sizeType& off)
      :_id(id),_stride(stride),_off(off) {}
    void operator++() {
      _id++;
    }
    bool operator!=(const IteratorIndex& other) const {
      return _id < other._id;
    }
    virtual ID operator*() const {
      ID ret;
      ret.resize(3);
      for(sizeType i=0; i<(sizeType)ret.size(); i++)ret[i]=(_stride == 0) ? _id+_off*i : _id*_stride+i;
      return ret;
    }
    sizeType _id;
    sizeType _stride;
    sizeType _off;
  };
public:
  VTKWriter(const string& name,const string& path,bool binary)
    :_os(path,binary ? ios_base::binary : ios_base::out),
     _points(binary ? ios_base::binary : ios_base::out),
     _cells(binary ? ios_base::binary : ios_base::out),
     _cellTypes(binary ? ios_base::binary : ios_base::out),
     _cellDatas(binary ? ios_base::binary : ios_base::out),
     _nrPoint(0),_nrCell(0),_nrIndex(0),_nrData(0),_vdt(UNSTRUCTURED_GRID),
     _binary(binary) {
    _os << "# vtk DataFile Version 1.0" << endl;
    _os << name << endl;
    _os << (binary ? "BINARY" : "ASCII") << endl;
    _os << "DATASET " << "UNSTRUCTURED_GRID" << endl;
  }
  virtual ~VTKWriter() {
    bool first;
    switch(_vdt) {
    case UNSTRUCTURED_GRID:
      _os << "POINTS " << _nrPoint << " " << (sizeof(T) == sizeof(float) ? "float" : "double") << endl;
      _os << _points.str();
      _os << "CELLS " << _nrCell << " " << _nrIndex << endl;
      _os << _cells.str();
      _os << "CELL_TYPES " << _nrCell << endl;
      _os << _cellTypes.str();
      first=false;
      for(typename map<string,Data>::const_iterator beg=_customData.begin(),end=_customData.end(); beg!=end; beg++) {
        if(!first)
          _os << "CELL_DATA " << beg->second._nr << endl;
        first=true;
        _os << beg->second._str << endl;
      }
      first=false;
      for(typename map<string,Data>::const_iterator beg=_customPointData.begin(),end=_customPointData.end(); beg!=end; beg++) {
        if(!first)
          _os << "POINT_DATA " << beg->second._nr << endl;
        first=true;
        _os << beg->second._str << endl;
      }
      break;
    case STRUCTURED_POINTS:
      ASSERT(_nrData == 1)
      _os << "POINT_DATA " << _nrPoint << std::endl;
      _os << "SCALARS data " << (sizeof(T) == sizeof(float) ? "float" : "double") << endl;
      _os << "LOOKUP_TABLE default" << endl;
      _os << _cellDatas.str();
      break;
    default:
      ASSERT_MSG(false,"Unsupported!")
    }
  }
  template <typename ITER> VTKWriter& appendPoints(ITER beg,ITER end) {
    typedef typename ValueTraits<ITER>::value_type value_type;
    sizeType nr=0;
    if(_binary) {
      for(; beg != end; ++beg) {
        const value_type& val=*beg;
        vtkWrite(_points,val[0]);
        vtkWrite(_points,val[1]);
        vtkWrite(_points,val[2]);
        nr++;
      }
    } else {
      for(; beg != end; ++beg) {
        const value_type& val=*beg;
        _points << (T)val[0] << " " << (T)val[1] << " " << (T)val[2] << endl;
        nr++;
      }
    }
    _nrPoint+=nr;
    return *this;
  }
  template <typename ITER> VTKWriter& appendCells(ITER beg,ITER end,VTK_CELL_TYPE ct,bool relativeIndex=false) {
    if(relativeIndex)
      ASSERT(_relativeCellIndex >= -1)
      int base=relativeIndex ? (int)_relativeCellIndex : 0;

    typedef typename ValueTraits<ITER>::value_type value_type;
    sizeType nr=0;
    sizeType nrIndex=0;
    if(_binary) {
      for(; beg != end; ++beg) {
        const value_type& val=*beg;
        switch(ct) {
        case POINT:
          nrIndex+=2;
          vtkWrite(_cells,1);
          vtkWrite(_cells,base+(int)val[0]);
          break;
        case LINE:
          nrIndex+=3;
          vtkWrite(_cells,2);
          vtkWrite(_cells,base+(int)val[0]);
          vtkWrite(_cells,base+(int)val[1]);
          break;
        case TRIANGLE:
        case QUADRATIC_LINE:
          nrIndex+=4;
          vtkWrite(_cells,3);
          vtkWrite(_cells,base+(int)val[0]);
          vtkWrite(_cells,base+(int)val[1]);
          vtkWrite(_cells,base+(int)val[2]);
          break;
        case TETRA:
        case QUAD:
          nrIndex+=5;
          vtkWrite(_cells,4);
          vtkWrite(_cells,base+(int)val[0]);
          vtkWrite(_cells,base+(int)val[1]);
          vtkWrite(_cells,base+(int)val[2]);
          vtkWrite(_cells,base+(int)val[3]);
          break;
        case VOXEL:
        case HEX:
          nrIndex+=9;
          vtkWrite(_cells,8);
          vtkWrite(_cells,base+(int)val[0]);
          vtkWrite(_cells,base+(int)val[1]);
          vtkWrite(_cells,base+(int)val[2]);
          vtkWrite(_cells,base+(int)val[3]);
          vtkWrite(_cells,base+(int)val[4]);
          vtkWrite(_cells,base+(int)val[5]);
          vtkWrite(_cells,base+(int)val[6]);
          vtkWrite(_cells,base+(int)val[7]);
          break;
        case POLYLINE:
          nrIndex+=val.size()+1;
          vtkWrite(_cells,(int)val.size());
          for(sizeType i=0; i<(int)val.size(); i++)
            vtkWrite(_cells,base+(int)val[i]);
          break;
        }
        vtkWrite(_cellTypes,(int)ct);
        nr++;
      }
    } else {
      for(; beg != end; ++beg) {
        const value_type& val=*beg;
        switch(ct) {
        case POINT:
          nrIndex+=2;
          _cells << "1 " << (base+(int)val[0]) << endl;
          break;
        case LINE:
          nrIndex+=3;
          _cells << "2 " << (base+(int)val[0]) << " " << (base+(int)val[1]) << endl;
          break;
        case TRIANGLE:
        case QUADRATIC_LINE:
          nrIndex+=4;
          _cells << "3 " << (base+(int)val[0]) << " " << (base+(int)val[1]) << " " << (base+(int)val[2]) << endl;
          break;
        case TETRA:
        case QUAD:
          nrIndex+=5;
          _cells << "4 " << (base+(int)val[0]) << " " << (base+(int)val[1]) << " " << (base+(int)val[2]) << " " << (base+(int)val[3]) << endl;
          break;
        case VOXEL:
        case HEX:
          nrIndex+=9;
          _cells << "8 " << (base+(int)val[0]) << " " << (base+(int)val[1]) << " " << (base+(int)val[2]) << " " << (base+(int)val[3]) << " "
                 << (base+(int)val[4]) << " " << (base+(int)val[5]) << " " << (base+(int)val[6]) << " " << (base+(int)val[7]) << endl;
          break;
        case POLYLINE:
          nrIndex+=val.size()+1;
          _cells << val.size() << " ";
          for(sizeType i=0; i<(int)val.size(); i++)
            _cells << (base+(int)val[i]) << " ";
          _cells << endl;
          break;
        }
        _cellTypes << ct << endl;
        nr++;
      }
    }
    _nrCell+=nr;
    _nrIndex+=nrIndex;
    return *this;
  }
  template <typename ITER> VTKWriter& appendDatas(const string name,ITER beg,ITER end) {
    if(_binary)
      for(; beg != end; ++beg,++_nrPoint)
        vtkWrite(_cellDatas,*beg);
    else
      for(; beg != end; ++beg,++_nrPoint)
        _cellDatas << (T)*beg << endl;
    _nrData++;
    return *this;
  }
  template <typename ITER> VTKWriter& appendCustomData(const string name,ITER beg,ITER end) {
    ostringstream os;
    if(_customData.find(name) == _customData.end()) {
      os << "SCALARS " << name << " " << (sizeof(T) == sizeof(float) ? "float" : "double") << endl;
      os << "LOOKUP_TABLE default" << endl;
    }

    Data& dat=_customData[name];
    if(_binary)
      for(; beg != end; ++beg,++dat._nr)
        vtkWrite(os,(T)*beg);
    else
      for(; beg != end; ++beg,++dat._nr)
        os << (T)*beg << endl;
    dat._str+=os.str();
    ASSERT(dat._nr == _nrCell)
    return *this;
  }
  template <typename ITER> VTKWriter& appendCustomPointData(const string name,ITER beg,ITER end) {
    ostringstream os;
    if(_customPointData.find(name) == _customPointData.end()) {
      //Data& dat=_customPointData[name];
      os << "SCALARS " << name << " " << (sizeof(T) == sizeof(float) ? "float" : "double") << endl;
      os << "LOOKUP_TABLE default" << endl;
    }

    Data& dat=_customPointData[name];
    if(_binary)
      for(; beg != end; ++beg,++dat._nr)
        vtkWrite(os,(T)*beg);
    else
      for(; beg != end; ++beg,++dat._nr)
        os << (T)*beg << endl;
    dat._str+=os.str();
    ASSERT(dat._nr == _nrPoint)
    return *this;
  }
  template <typename ITER> VTKWriter& appendCustomVectorData(const string name,ITER beg,ITER end) {
    ostringstream os;
    if(_customData.find(name) == _customData.end()) {
      os << "VECTORS " << name << " " << (sizeof(T) == sizeof(float) ? "float" : "double") << endl;
    }

    Data& dat=_customData[name];
    if(_binary)
      for(; beg != end; ++beg,++dat._nr) {
        vtkWrite(os,(T)(*beg)[0]);
        vtkWrite(os,(T)(*beg)[1]);
        vtkWrite(os,(T)(*beg)[2]);
      }
    else
      for(; beg != end; ++beg,++dat._nr)
        os << (T)(*beg)[0] << " " << (T)(*beg)[1] << " " << (T)(*beg)[2] << endl;
    dat._str+=os.str();
    ASSERT(dat._nr == _nrCell)
    return *this;
  }
  template <typename ITER> VTKWriter& appendCustomPointVectorData(const string name,ITER beg,ITER end) {
    ostringstream os;
    if(_customPointData.find(name) == _customPointData.end()) {
      //Data& dat=_customPointData[name];
      os << "VECTORS " << name << " " << (sizeof(T) == sizeof(float) ? "float" : "double") << endl;
    }

    Data& dat=_customPointData[name];
    if(_binary)
      for(; beg != end; ++beg,++dat._nr) {
        vtkWrite(os,(T)(*beg)[0]);
        vtkWrite(os,(T)(*beg)[1]);
        vtkWrite(os,(T)(*beg)[2]);
      }
    else
      for(; beg != end; ++beg,++dat._nr)
        os << (T)(*beg)[0] << " " << (T)(*beg)[1] << " " << (T)(*beg)[2] << endl;
    dat._str+=os.str();
    ASSERT(dat._nr == _nrPoint)
    return *this;
  }
  template <typename ITER> VTKWriter& appendCustomPointColorData(const string name,ITER begC,ITER endC) {
    ostringstream os;
    if(_customPointData.find(name) == _customPointData.end()) {
      //Data& dat=_customPointData[name];
      os << "COLOR_SCALARS " << name << " " << (*begC).size() << endl;
    }

    int sz=(int)(*begC).size();
    Data& dat=_customPointData[name];
    if(_binary) {
      for(; begC != endC; ++begC,++dat._nr)
        for(int d=0; d<sz; d++)
          vtkWrite(os,(unsigned char)((*begC)[d]*255));
    } else {
      for(; begC != endC; ++begC,++dat._nr) {
        for(int d=0; d<sz; d++)
          os << (T)(*begC)[d] << " ";
        os << endl;
      }
    }
    dat._str+=os.str();
    return *this;
  }
  void setRelativeIndex(sizeType rel=-1) {
    if(rel == -1)
      _relativeCellIndex=_nrPoint;
    else _relativeCellIndex=rel;
  }
private:
  ofstream _os;
  ostringstream _points,_cells,_cellTypes,_cellDatas;
  map<string,Data> _customData;
  map<string,Data> _customPointData;
  sizeType _nrPoint,_nrCell,_nrIndex,_nrData;
  VTK_DATA_TYPE _vdt;
  bool _binary;
  sizeType _relativeCellIndex;
};

//reader
void readCurve(istream& is,QuadraticCurveSet& curves)
{
  sizeType n,tmp;
  is.read((char*)&n,sizeof(sizeType));
  curves._curves.resize(n);
  for(sizeType i=0;i<n;i++) {
    is.read((char*)&tmp,sizeof(sizeType));
    is.read((char*)&tmp,sizeof(sizeType));
    is.read((char*)&curves._curves[i]._coef[0][0],sizeof(scalar));
    is.read((char*)&curves._curves[i]._coef[0][1],sizeof(scalar));
    is.read((char*)&curves._curves[i]._coef[0][2],sizeof(scalar));
    is.read((char*)&curves._curves[i]._coef[1][0],sizeof(scalar));
    is.read((char*)&curves._curves[i]._coef[1][1],sizeof(scalar));
    is.read((char*)&curves._curves[i]._coef[1][2],sizeof(scalar));
    is.read((char*)&curves._curves[i]._coef[2][0],sizeof(scalar));
    is.read((char*)&curves._curves[i]._coef[2][1],sizeof(scalar));
    is.read((char*)&curves._curves[i]._coef[2][2],sizeof(scalar));
  }
  is.read((char*)&curves._nrCandidate,sizeof(sizeType));
  is.read((char*)&curves._time0,sizeof(scalar));
}
void readVec(istream& is,vector<scalar>& vecs)
{
  sizeType n,tmp;
  is.read((char*)&n,sizeof(sizeType));
  is.read((char*)&tmp,sizeof(sizeType));
  vecs.resize(n);
  for(sizeType i=0;i<n;i++)
    is.read((char*)&vecs[i],sizeof(scalar));
}
vector<scalar> vec(scalar a,scalar b,scalar c=0)
{
  vector<scalar> v;
  v.push_back(a);
  v.push_back(b);
  v.push_back(c);
  return v;
}
int main(int argc,char** argv)
{
  ifstream is(argv[1],ios::binary);
  FluidRLData data;

  sizeType n;
  //quadratic curves
  is.read((char*)&n,sizeof(sizeType));
  data._curves.resize(n);
  for(sizeType i=0;i<n;i++)
    readCurve(is,data._curves[i]);
  //state/observation
  is.read((char*)&n,sizeof(sizeType));
  data._features.resize(n);
  for(sizeType i=0;i<n;i++)
    readVec(is,data._features[i]);
  //action label
  is.read((char*)&n,sizeof(sizeType));
  data._actions.resize(n);
  for(sizeType i=0;i<n;i++)
    readVec(is,data._actions[i]);
  //visualize
  for(sizeType f=0;f<n;f++) {
    char buf[65536];
    sprintf(buf,"./frame%ld.vtk",f);
    VTKWriter<scalar> os("frameData",buf,true);
    vector<vector<scalar> > vss,css;
    vector<scalar> feature=data._features[f];
    //draw source container speed
    vss.push_back(vec(feature[0],feature[1]));
    vss.push_back(vec(feature[0]+feature[2],feature[1]+feature[3]));
    css.push_back(vec(0,1,0));
    css.push_back(vec(0,0,0));
    //draw source container desired speed
    //vss.push_back(vec(feature[0],feature[1]));
    //vss.push_back(vec(feature[0]+action[0],feature[1]+action[1]));
    //css.push_back(vec(0,1,0));
    //css.push_back(vec(0,0,0));
    //draw target container speed
    vss.push_back(vec(feature[4],feature[5]));
    vss.push_back(vec(feature[4]+feature[6],feature[5]+feature[7]));
    css.push_back(vec(1,0,0));
    css.push_back(vec(0,0,0));
    //draw fluid height
    sizeType nrSample=((sizeType)feature.size()-9)/3;
    for(sizeType i=0;i<nrSample;i++) {
      vss.push_back(vec(feature[9+i*3],feature[9+i*3+2]));
      vss.push_back(vec(feature[9+i*3],feature[9+i*3+1]));
      css.push_back(vec(0.3,0.6,0.9));
      css.push_back(vec(0.3,0.6,0.9));
      if(i > 0) {
        vss.push_back(vec(feature[9+(i-1)*3],feature[9+(i-1)*3+2]));
        vss.push_back(vec(feature[9+i*3],feature[9+i*3+2]));
        css.push_back(vec(0.3,0.6,0.9));
        css.push_back(vec(0.3,0.6,0.9));
      }
    }
    //draw rigid height
    for(sizeType i=1;i<nrSample;i++) {
      vss.push_back(vec(feature[9+(i-1)*3],feature[9+(i-1)*3+1]));
      vss.push_back(vec(feature[9+i*3],feature[9+i*3+1]));
      css.push_back(vec(0,1,0));
      css.push_back(vec(0,1,0));
    }
    //draw quadratic curves
    scalar x,y,tc;
    const QuadraticCurveSet& curves=data._curves[f];
    for(sizeType i=0;i<(sizeType)curves._curves.size();i++) {
      QuadraticCurve c=curves._curves[i];
      for(scalar t=curves._time0,t1=t+0.2f;t<t1;t+=0.01f) {
        tc=t-curves._time0;
        x=c._coef[0][0]*tc*tc+c._coef[0][1]*tc+c._coef[0][2];
        y=c._coef[1][0]*tc*tc+c._coef[1][1]*tc+c._coef[1][2];
        vss.push_back(vec(x,y));
        tc=t-curves._time0+0.01f;
        x=c._coef[0][0]*tc*tc+c._coef[0][1]*tc+c._coef[0][2];
        y=c._coef[1][0]*tc*tc+c._coef[1][1]*tc+c._coef[1][2];
        vss.push_back(vec(x,y));
        css.push_back(vec(0.3f,0.6f,0.9f));
        css.push_back(vec(0.3f,0.6f,0.9f));
      }
    }
    //output
    os.appendPoints(vss.begin(),vss.end());
    os.appendCells(VTKWriter<scalar>::IteratorIndex<vector<sizeType> >(0,2,0),
                   VTKWriter<scalar>::IteratorIndex<vector<sizeType> >((sizeType)vss.size()/2,2,0),
                   VTKWriter<scalar>::LINE);
    os.appendCustomPointColorData("Color",css.begin(),css.end());
  }
  return 0;
}
